#!/usr/bin/env perl

use strict;
use warnings;
use IO::File;
use Fcntl;

my $cmdname = $0; $cmdname =~ s{.*/}{};
my $usage = "Usage: $cmdname graph.txt graph.mp\n";
die $usage if @ARGV != 2;
my $infile = $ARGV[0];
my $outfile = $ARGV[1];

# first scan the input file to determine our penalty limit
my $penalty_limit;
my $in = new IO::File $infile
   or die "$cmdname: unable to open $infile: $!\n";
my @badness = ();
while (<$in>) {
   chomp;
   if (m{^F\|}) {
      @badness = (); next;
   }
   if (m{^C\|\d+\|.*?\|(\d+)}) {
      push(@badness, $1);
   }
}
@badness = sort { $a <=> $b } @badness;
if (scalar @badness <= 5) {
   $penalty_limit = pop(@badness);
} else {
   $penalty_limit = $badness[5];
}
die "no penalty limit" unless defined $penalty_limit;
print "penalty limit = $penalty_limit\n";

$in->seek(0, 0) or die "$cmdname: unable to rewind $infile: $!\n";

my $id = 0;
my %backlinks;
my %word;
$word{$id} = "";
my %level;
$level{$id} = 0;
my %successor;
my %penalty;
$penalty{$id} = 0;
my %min_predecessor_penalty;
my %best_predecessor;
my %opt;
my $lastid;
while (<$in>) {
   chomp;
   my ($key, @fields) = split /\|/;
   if ($key eq "W") {
      my ($bpid, $bpword) = @fields;
      $id = $bpid;
      $word{$id} = $bpword;
   } elsif ($key eq "C") {
      my ($bpid, $bpword, $penalty) = @fields;
      unless (defined $penalty) {
	 die "bad input: $_\n";
      }
      if (defined $penalty{$id}) {
	 if ($penalty < $penalty{$id}) {
	    $penalty{$id} = $penalty;
	 }
      } else {
	 $penalty{$id} = $penalty;
      }
      next if $penalty == 1073741823;
      next if $penalty > $penalty_limit;
      if (!defined($min_predecessor_penalty{$id}) ||
	    $penalty < $min_predecessor_penalty{$id}) {
	 $min_predecessor_penalty{$id} = $penalty;
	 $best_predecessor{$id} = $bpid;
      }
      $penalty -= $penalty{$bpid};
      $backlinks{$id}->{$bpid} = $penalty;
      $successor{$bpid}->{$id} = $penalty;
      my $newlevel = $level{$bpid} + 1;
      if (defined $level{$id}) {
	 if ($level{$id} < $newlevel) {
	    $level{$id} = $newlevel;
	 }
      } else {
	 $level{$id} = $newlevel;
      }
   } elsif ($key eq "F") {
      # my $bpword;
      # ($id, $bpword) = @fields;
      # $word{$id} = $bpword;
      $lastid = $id;
   }
}

$id = $lastid;
$opt{$id} = 1;
while (defined $best_predecessor{$id}) {
   $id = $best_predecessor{$id};
   $opt{$id} = 1;
}

my %included;
my @worklist = ($lastid);
while (@worklist > 0) {
   my $id = shift @worklist;
   next if $included{$id};
   $included{$id} = 1;
   # die "$cmdname penalty limit is too small ($penalty_limit)\n" unless defined $level{$id};
   foreach my $i (keys %{$backlinks{$id}}) {
      push(@worklist, $i);
   }
}

my $out = new IO::File $outfile, O_WRONLY|O_CREAT|O_TRUNC, 0600
   or die "$cmdname: unable to create $outfile: $!\n";

my $headin = new IO::File "graph.head" or die;
print $out $_ while (<$headin>);
$headin->close;

my @arrows = ();
my %boxname;
my $boxid = "a"; my $lastlevel; my $lastbox; my $firstbox;
for my $id (sort { $level{$a} <=> $level{$b} || $a <=> $b } keys %included) {
   my @successors = ();
   foreach my $successor (sort keys %{$successor{$id}}) {
      push(@successors, $successor) if defined $included{$successor};
   }
   my $level = $level{$id};
   next unless defined $level;
   my $word = $word{$id};
   die "bad word for id = $id" unless defined $word;
   my $boxname = getname($id);
   print $out "Node(", $boxname, ", btex ", $word, " etex, ", $level, ");\n";
   if (defined $lastlevel) {
      if ($lastlevel == $level) {
	 print $out "Join(", $lastbox, ", ", $boxname, ");\n";
      } else {
	 if (defined $lastbox) {
	    print $out "Center($firstbox, $lastbox);\n";
	 }
	 $firstbox = $boxname;
      }
   } elsif (!defined $firstbox) {
      $firstbox = $boxname;
   }
   foreach my $successor (@successors) {
      next if $successor == $id;
      my $penalty = $successor{$id}->{$successor};
      push(@arrows, {
	 from => $boxname,
	 to => getname($successor),
	 penalty => $penalty,
	 best => defined $opt{$successor} &&
		     $best_predecessor{$successor} == $id,
      });
   }
   $lastlevel = $level; $lastbox = $boxname;
}
if (defined $lastbox) {
   print $out "Center($firstbox, $lastbox);\n";
}
print $out "beginfig(1);\n";
print $out "   boxheight = 2 mbh;\n";
foreach my $arrow (@arrows) {
   if ($arrow->{best}) {
      print $out "   RedArrow(";
   } else {
      print $out "   Arrow(";
   }
   die "murks" if $arrow->{from} eq $arrow->{to};
   print $out $arrow->{from}, ", ", $arrow->{to},
      ", btex {\\scriptsize ", $arrow->{penalty}, "} etex);\n";
}
print $out "endfig;\nend.\n";
$out->close;

sub getname {
   my ($id) = @_;
   return $boxname{$id} if defined $boxname{$id};
   my $boxname = "Box" . $boxid++;
   $boxname{$id} = $boxname;
   return $boxname;
}
